"use strict";
var __createBinding = (this && this.__createBinding) || (Object.create ? (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    Object.defineProperty(o, k2, { enumerable: true, get: function() { return m[k]; } });
}) : (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    o[k2] = m[k];
}));
var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function(o, v) {
    Object.defineProperty(o, "default", { enumerable: true, value: v });
}) : function(o, v) {
    o["default"] = v;
});
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k)) __createBinding(result, mod, k);
    __setModuleDefault(result, mod);
    return result;
};
var __spreadArrays = (this && this.__spreadArrays) || function () {
    for (var s = 0, i = 0, il = arguments.length; i < il; i++) s += arguments[i].length;
    for (var r = Array(s), k = 0, i = 0; i < il; i++)
        for (var a = arguments[i], j = 0, jl = a.length; j < jl; j++, k++)
            r[k] = a[j];
    return r;
};
exports.__esModule = true;
exports.plugin = void 0;
var graphql = __importStar(require("graphql"));
var prettier_1 = require("prettier");
var change_case_1 = require("change-case");
exports.plugin = function (schema, documents, config) {
    if (config === void 0) { config = {}; }
    var queryObj = schema.getType("Query");
    if (!graphql.isObjectType(queryObj)) {
        throw new Error("Unexpected error");
    }
    var scalarDecoders = config.scalarDecoders || {};
    var scalarDecoderImpls = Object.entries(scalarDecoders).map(function (_a) {
        var key = _a[0], val = _a[1];
        return "export const " + key + "Decoder = " + val;
    });
    var queryDecoderImpls = documents
        .flatMap(function (doc) { return doc.document.definitions; })
        .filter(isOperationDefinitionNode)
        .map(function (def) { return parseOperationDefinitionNode(def, queryObj); });
    var rawContent = __spreadArrays(scalarDecoderImpls, queryDecoderImpls).join("\n");
    var content = prettier_1.format(rawContent, { parser: "typescript" });
    return {
        prepend: ['import * as D from "@mojotech/json-type-validation";'],
        content: content
    };
};
var isOperationDefinitionNode = function (defNode) {
    return defNode.kind === "OperationDefinition";
};
// ## Example
// `
// export const TopicsQueryDecoder: D.Decoder<TopicsQuery> = D.object({ ... });
// `
var parseOperationDefinitionNode = function (node, object) {
    if (node.operation === "query") {
        var name_1 = change_case_1.pascalCase(node.name.value) + "Query";
        var decoderImpl = parseObjectInner(node.selectionSet, object);
        return "export const " + name_1 + "Decoder: D.Decoder<" + name_1 + "> = " + decoderImpl;
    }
    throw new Error("Unsupported operation type: " + node.operation);
};
// ## Example
// `
// D.string()
// `
//
// `
// D.array( D.union( D.constant(null), D.string() ) )
// `
var parseAny = function (node, type) {
    if (graphql.isNonNullType(type)) {
        var ofType = type.ofType;
        // scalar
        if (graphql.isScalarType(ofType)) {
            return parseScalar(ofType);
            // object
        }
        else if (graphql.isObjectType(ofType)) {
            return parseObject(node, ofType);
            // list
        }
        else if (graphql.isListType(ofType)) {
            return parseList(node, ofType);
            // enum
        }
        else if (graphql.isEnumType(ofType)) {
            return parseEnum(node, ofType);
        }
        // nullable scalar
    }
    else if (graphql.isScalarType(type)) {
        return parseNullable(parseScalar(type));
        // nullable object
    }
    else if (graphql.isObjectType(type)) {
        return parseNullable(parseObject(node, type));
        // nullable list
    }
    else if (graphql.isListType(type)) {
        return parseNullable(parseList(node, type));
    }
    // others
    throw new Error("unsupported type: " + type);
};
var parseNullable = function (parsed) {
    return "D.union( D.constant(null), " + parsed + " )";
};
var parseScalar = function (scalar) {
    switch (scalar.name) {
        case "Boolean":
            return "D.boolean()";
        case "ID":
        case "String":
            return "D.string()";
        case "Int":
        case "Float":
            return "D.number()";
        default:
            return scalar.name + "Decoder";
    }
};
var parseObject = function (node, object) {
    return parseObjectInner(node.selectionSet, object);
};
// # Return
// string such as
// `
// D.object({
//  id: UUIDDecoder,
//  title: StringDecoder,
// })
// `
var parseObjectInner = function (node, object) {
    var internal = node.selections
        .map(function (childNode) {
        if (childNode.kind !== "Field") {
            throw new Error("Unsupported selection type: " + childNode.kind);
        }
        var name = childNode.name.value;
        var type = object.getFields()[name].type;
        return name + ": " + parseAny(childNode, type);
    })
        .join(", ");
    return "D.object({ " + internal + " })";
};
var parseList = function (node, list) {
    return "D.array( " + parseAny(node, list.ofType) + " )";
};
// ## Example
// `
// D.oneOf( D.constant(Status.Open), D.constant(Status.Closed) )
// `
var parseEnum = function (node, enumType) {
    var name = enumType.name;
    var enumConsts = enumType
        .getValues()
        .map(function (val) { return "D.constant(" + name + "." + change_case_1.pascalCase(val.name) + ")"; })
        .join(", ");
    return "D.oneOf( " + enumConsts + " )";
};
